#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='trainingtorchsample',
      version='0.1.3',
      description='High-Level Training, Augmentation, and Sampling for Pytorch',
      author='NC Cullen',
      author_email='nickmarch31@yahoo.com',
      packages=find_packages()
     )